from __future__ import annotations

from enum import Enum


class SegmentCategory(Enum):
    """
    Segment category
    """

    Sponsor = "sponsor"
    Selfpromo = "selfpromo"
    Interaction = "interaction"
    Intro = "intro"
    Outro = "outro"
    Preview = "preview"
    MusicOfftopic = "music_offtopic"
    Filler = "filler"

    @classmethod
    def get_category(cls, segment_type: str) -> SegmentCategory:
        """
        Get segment category
        """
        if segment_type == "intro":
            return SegmentCategory.Intro
        elif segment_type == "outro":
            return SegmentCategory.Outro
        elif segment_type == "preview":
            return SegmentCategory.Preview
        elif segment_type == "music_offtopic":
            return SegmentCategory.MusicOfftopic
        elif segment_type == "filler":
            return SegmentCategory.Filler
        elif segment_type == "sponsor":
            return SegmentCategory.Sponsor
        elif segment_type == "selfpromo":
            return SegmentCategory.Selfpromo
        elif segment_type == "interaction":
            return SegmentCategory.Interaction
        else:
            raise ValueError(f"Unknown segment type: {segment_type}")

    @classmethod
    def get_category_name(cls, segment_type: str) -> str:
        """
        Get segment category name
        """
        return cls.get_category(segment_type).name

    @classmethod
    def get_category_from_name(cls, category_name: str) -> SegmentCategory:
        """
        Get segment category from name
        """
        return SegmentCategory[category_name]

    @classmethod
    def get_category_list(cls) -> list[SegmentCategory]:
        """
        Get segment category list
        """
        return list(cls)

    @classmethod
    def get_category_list_name(cls) -> list[str]:
        """
        Get segment category list name
        """
        return [category.name for category in cls]

    @classmethod
    def get_category_list_value(cls) -> list[str]:
        """
        Get segment category list value
        """
        return [category.value for category in cls]  # type: ignore



from __future__ import annotations

from piccolo.columns import JSONB, BigInt, Boolean, Text, Timestamptz
from piccolo.table import Table

from pylav.constants.config import CONFIG_DIR, USE_BUNDLED_EXTERNAL_LAVA_LINK_NODE, USE_BUNDLED_EXTERNAL_PYLAV_NODE
from pylav.storage.database.tables.misc import DATABASE_ENGINE


class LibConfigRow(Table, db=DATABASE_ENGINE, tablename="lib_config"):
    id = BigInt(index=True)
    bot = BigInt(primary_key=True, index=True)
    config_folder = Text(null=False, default=str(CONFIG_DIR))
    java_path = Text(null=False, default="java")
    enable_managed_node = Boolean(null=False, default=True)
    auto_update_managed_nodes = Boolean(null=False, default=True)
    localtrack_folder = Text(null=True, default=str(CONFIG_DIR / "music"))
    download_id = BigInt(index=True, default=0)
    update_bot_activity = Boolean(null=False, default=False)
    use_bundled_pylav_external = Boolean(null=False, default=USE_BUNDLED_EXTERNAL_PYLAV_NODE)
    use_bundled_lava_link_external = Boolean(null=False, default=USE_BUNDLED_EXTERNAL_LAVA_LINK_NODE)

    extras: JSONB = JSONB(null=False, default={})
    next_execution_update_bundled_playlists = Timestamptz(null=True, default=None)
    next_execution_update_bundled_external_playlists = Timestamptz(null=True, default=None)
    next_execution_update_external_playlists = Timestamptz(null=True, default=None)



from __future__ import annotations

import dataclasses
from typing import Annotated, Union

from pylav.nodes.api.responses.filters.misc import ValueRange


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class EqualizerBand:
    band: Annotated[int | None, ValueRange(min=0, max=14)]
    gain: Annotated[float | None, ValueRange(min=-0.25, max=1.0)]

    def to_dict(self) -> dict[str, float | int]:
        response: dict[str, float | int] = {}
        if self.band is not None:
            response["band"] = self.band
        if self.gain is not None:
            response["gain"] = self.gain
        return response


Equalizer = list[Union[EqualizerBand, dict]]



from __future__ import annotations

from collections.abc import AsyncIterator
from typing import TYPE_CHECKING

import discord
from piccolo.columns import Float

from pylav.exceptions.database import EntryNotFoundException
from pylav.logging import getLogger
from pylav.storage.database.tables.equalizer import EqualizerRow
from pylav.storage.models import equilizer
from pylav.type_hints.bot import DISCORD_BOT_TYPE
from pylav.utils.vendor.redbot import AsyncIter

LOGGER = getLogger("PyLav.Database.Controller.Equalizer")

if TYPE_CHECKING:
    from pylav.core.client import Client


class EqualizerController:
    __slots__ = ("_client",)

    def __init__(self, client: Client) -> None:
        self._client = client

    @property
    def client(self) -> Client:
        return self._client

    @staticmethod
    async def get_equalizer_by_name(equalizer_name: str, limit: int = None) -> list[equilizer.Equalizer]:
        if limit is None:
            equalizers = (
                await EqualizerRow.select()
                .where(EqualizerRow.name.ilike(f"%{equalizer_name.lower()}%"))
                .output(load_json=True, nested=True)
            )
        else:
            equalizers = (
                await EqualizerRow.select().where(EqualizerRow.name.ilike(f"%{equalizer_name.lower()}%")).limit(limit)
            ).output(load_json=True, nested=True)

        if not equalizers:
            raise EntryNotFoundException(f"Equalizer with name {equalizer_name} not found")
        return [equilizer.Equalizer(**equalizer) for equalizer in equalizers]

    @staticmethod
    async def get_equalizer_by_id(identifier: int | str) -> equilizer.Equalizer:
        try:
            identifier = int(identifier)
        except ValueError as e:
            raise EntryNotFoundException(f"Equalizer with id {identifier} not found") from e
        equalizer = (
            await EqualizerRow.select().where(EqualizerRow.id == identifier).first().output(load_json=True, nested=True)
        )
        if not equalizer:
            raise EntryNotFoundException(f"Equalizer with ID {identifier} not found")
        return equilizer.Equalizer(**equalizer)

    async def get_equalizer_by_name_or_id(
        self, equalizer_name_or_id: int | str, limit: int = None
    ) -> list[equilizer.Equalizer]:
        try:
            return [await self.get_equalizer_by_id(equalizer_name_or_id)]
        except EntryNotFoundException:
            return await self.get_equalizer_by_name(equalizer_name_or_id, limit=limit)

    @staticmethod
    async def get_equalizers_by_author(author: int) -> list[equilizer.Equalizer]:
        equalizers = (
            await EqualizerRow.select().where(EqualizerRow.author == author).output(load_json=True, nested=True)
        )
        if not equalizers:
            raise EntryNotFoundException(f"Equalizer with author {author} not found")
        return [equilizer.Equalizer(**equalizer) for equalizer in equalizers]

    @staticmethod
    async def get_equalizers_by_scope(scope: int) -> list[equilizer.Equalizer]:
        equalizers = await EqualizerRow.select().where(EqualizerRow.scope == scope).output(load_json=True, nested=True)
        if not equalizers:
            raise EntryNotFoundException(f"Equalizer with scope {scope} not found")
        return [equilizer.Equalizer(**equalizer) for equalizer in equalizers]

    @staticmethod
    async def get_all_equalizers() -> AsyncIterator[equilizer.Equalizer]:
        for entry in await EqualizerRow.select().output(load_json=True, nested=True):
            yield equilizer.Equalizer(**entry)

    @staticmethod
    def _get_equalizer_band_defaults(*args: tuple[Float, float | None]) -> dict[Float, float]:
        null_values = {None, 0.0}
        return {band: value for band, value in args if value not in null_values}

    @staticmethod
    async def create_or_update_equalizer(
        identifier: int,
        scope: int,
        author: int,
        name: str,
        description: str | None = None,
        band_25: float | None = None,
        band_40: float | None = None,
        band_63: float | None = None,
        band_100: float | None = None,
        band_160: float | None = None,
        band_250: float | None = None,
        band_400: float | None = None,
        band_630: float | None = None,
        band_1000: float | None = None,
        band_1600: float | None = None,
        band_2500: float | None = None,
        band_4000: float | None = None,
        band_6300: float | None = None,
        band_10000: float | None = None,
        band_16000: float | None = None,
    ) -> equilizer.Equalizer:
        values = {
            EqualizerRow.scope: scope,
            EqualizerRow.author: author,
            EqualizerRow.name: name,
            EqualizerRow.description: description,
        }
        values |= EqualizerController._get_equalizer_band_defaults(
            (EqualizerRow.band_25, band_25),
            (EqualizerRow.band_40, band_40),
            (EqualizerRow.band_63, band_63),
            (EqualizerRow.band_100, band_100),
            (EqualizerRow.band_160, band_160),
            (EqualizerRow.band_250, band_250),
            (EqualizerRow.band_400, band_400),
            (EqualizerRow.band_630, band_630),
            (EqualizerRow.band_1000, band_1000),
            (EqualizerRow.band_1600, band_1600),
            (EqualizerRow.band_2500, band_2500),
            (EqualizerRow.band_4000, band_4000),
            (EqualizerRow.band_6300, band_6300),
            (EqualizerRow.band_10000, band_10000),
            (EqualizerRow.band_16000, band_16000),
        )
        equalizer = (
            await EqualizerRow.objects()
            .output(load_json=True)
            .get_or_create(EqualizerRow.id == identifier, defaults=values)
        )
        # noinspection PyProtectedMember
        if not equalizer._was_created:
            await EqualizerRow.update(values).where(EqualizerRow.id == identifier)
        return equilizer.Equalizer(**equalizer.to_dict())

    @staticmethod
    async def delete_equalizer(equalizer_id: int) -> None:
        await EqualizerRow.delete().where(EqualizerRow.id == equalizer_id)

    @staticmethod
    async def get_all_equalizers_by_author(author: int) -> AsyncIterator[equilizer.Equalizer]:
        for entry in (
            await EqualizerRow.select().where(EqualizerRow.author == author).output(load_json=True, nested=True)
        ):
            yield equilizer.Equalizer(**entry)

    @staticmethod
    async def get_all_equalizers_by_scope(scope: int) -> AsyncIterator[equilizer.Equalizer]:
        for entry in await EqualizerRow.select().where(EqualizerRow.scope == scope).output(load_json=True, nested=True):
            yield equilizer.Equalizer(**entry)

    @staticmethod
    async def get_all_equalizers_by_scope_and_author(scope: int, author: int) -> AsyncIterator[equilizer.Equalizer]:
        for entry in (
            await EqualizerRow.select()
            .where(
                EqualizerRow.scope == scope,
                EqualizerRow.author == author,
            )
            .output(load_json=True, nested=True)
        ):
            yield equilizer.Equalizer(**entry)

    async def get_global_equalizers(self) -> AsyncIterator[equilizer.Equalizer]:
        for entry in (
            await EqualizerRow.select()
            .where(EqualizerRow.scope == self._client.bot.user.id)  # type: ignore
            .output(load_json=True, nested=True)
        ):
            yield equilizer.Equalizer(**entry)

    async def create_or_update_global_equalizer(
        self,
        identifier: int,
        author: int,
        name: str,
        description: str | None = None,
        band_25: float | None = None,
        band_40: float | None = None,
        band_63: float | None = None,
        band_100: float | None = None,
        band_160: float | None = None,
        band_250: float | None = None,
        band_400: float | None = None,
        band_630: float | None = None,
        band_1000: float | None = None,
        band_1600: float | None = None,
        band_2500: float | None = None,
        band_4000: float | None = None,
        band_6300: float | None = None,
        band_10000: float | None = None,
        band_16000: float | None = None,
    ) -> equilizer.Equalizer:
        return await self.create_or_update_equalizer(
            identifier=identifier,
            scope=self._client.bot.user.id,
            author=author,
            name=name,
            description=description,
            band_25=band_25,
            band_40=band_40,
            band_63=band_63,
            band_100=band_100,
            band_160=band_160,
            band_250=band_250,
            band_400=band_400,
            band_630=band_630,
            band_1000=band_1000,
            band_1600=band_1600,
            band_2500=band_2500,
            band_4000=band_4000,
            band_6300=band_6300,
            band_10000=band_10000,
            band_16000=band_16000,
        )

    async def create_or_update_user_equalizer(
        self,
        user_id: int,
        author: int,
        name: str,
        description: str | None = None,
        band_25: float | None = None,
        band_40: float | None = None,
        band_63: float | None = None,
        band_100: float | None = None,
        band_160: float | None = None,
        band_250: float | None = None,
        band_400: float | None = None,
        band_630: float | None = None,
        band_1000: float | None = None,
        band_1600: float | None = None,
        band_2500: float | None = None,
        band_4000: float | None = None,
        band_6300: float | None = None,
        band_10000: float | None = None,
        band_16000: float | None = None,
    ) -> equilizer.Equalizer:
        return await self.create_or_update_equalizer(
            identifier=user_id,
            scope=author,
            author=author,
            name=name,
            description=description,
            band_25=band_25,
            band_40=band_40,
            band_63=band_63,
            band_100=band_100,
            band_160=band_160,
            band_250=band_250,
            band_400=band_400,
            band_630=band_630,
            band_1000=band_1000,
            band_1600=band_1600,
            band_2500=band_2500,
            band_4000=band_4000,
            band_6300=band_6300,
            band_10000=band_10000,
            band_16000=band_16000,
        )

    async def create_or_update_channel_equalizer(
        self,
        text_channel: discord.abc.MessageableChannel,
        identifier: int,
        author: int,
        name: str,
        description: str | None = None,
        band_25: float | None = None,
        band_40: float | None = None,
        band_63: float | None = None,
        band_100: float | None = None,
        band_160: float | None = None,
        band_250: float | None = None,
        band_400: float | None = None,
        band_630: float | None = None,
        band_1000: float | None = None,
        band_1600: float | None = None,
        band_2500: float | None = None,
        band_4000: float | None = None,
        band_6300: float | None = None,
        band_10000: float | None = None,
        band_16000: float | None = None,
    ) -> equilizer.Equalizer:
        return await self.create_or_update_equalizer(
            identifier=identifier,
            scope=text_channel.id,
            author=author,
            name=name,
            description=description,
            band_25=band_25,
            band_40=band_40,
            band_63=band_63,
            band_100=band_100,
            band_160=band_160,
            band_250=band_250,
            band_400=band_400,
            band_630=band_630,
            band_1000=band_1000,
            band_1600=band_1600,
            band_2500=band_2500,
            band_4000=band_4000,
            band_6300=band_6300,
            band_10000=band_10000,
            band_16000=band_16000,
        )

    async def create_or_update_guild_equalizer(
        self,
        guild: discord.Guild,
        identifier: int,
        author: int,
        name: str,
        description: str | None = None,
        band_25: float | None = None,
        band_40: float | None = None,
        band_63: float | None = None,
        band_100: float | None = None,
        band_160: float | None = None,
        band_250: float | None = None,
        band_400: float | None = None,
        band_630: float | None = None,
        band_1000: float | None = None,
        band_1600: float | None = None,
        band_2500: float | None = None,
        band_4000: float | None = None,
        band_6300: float | None = None,
        band_10000: float | None = None,
        band_16000: float | None = None,
    ) -> equilizer.Equalizer:
        return await self.create_or_update_equalizer(
            identifier=identifier,
            scope=guild.id,
            author=author,
            name=name,
            description=description,
            band_25=band_25,
            band_40=band_40,
            band_63=band_63,
            band_100=band_100,
            band_160=band_160,
            band_250=band_250,
            band_400=band_400,
            band_630=band_630,
            band_1000=band_1000,
            band_1600=band_1600,
            band_2500=band_2500,
            band_4000=band_4000,
            band_6300=band_6300,
            band_10000=band_10000,
            band_16000=band_16000,
        )

    async def create_or_update_vc_equalizer(
        self,
        voice_channel: discord.channel.VocalGuildChannel,
        identifier: int,
        author: int,
        name: str,
        description: str | None = None,
        band_25: float | None = None,
        band_40: float | None = None,
        band_63: float | None = None,
        band_100: float | None = None,
        band_160: float | None = None,
        band_250: float | None = None,
        band_400: float | None = None,
        band_630: float | None = None,
        band_1000: float | None = None,
        band_1600: float | None = None,
        band_2500: float | None = None,
        band_4000: float | None = None,
        band_6300: float | None = None,
        band_10000: float | None = None,
        band_16000: float | None = None,
    ) -> equilizer.Equalizer:
        return await self.create_or_update_equalizer(
            identifier=identifier,
            scope=voice_channel.id,
            author=author,
            name=name,
            description=description,
            band_25=band_25,
            band_40=band_40,
            band_63=band_63,
            band_100=band_100,
            band_160=band_160,
            band_250=band_250,
            band_400=band_400,
            band_630=band_630,
            band_1000=band_1000,
            band_1600=band_1600,
            band_2500=band_2500,
            band_4000=band_4000,
            band_6300=band_6300,
            band_10000=band_10000,
            band_16000=band_16000,
        )

    async def get_all_for_user(
        self,
        requester: int,
        *,
        voice_channel: discord.channel.VocalGuildChannel = None,
        guild: discord.Guild = None,
        channel: discord.abc.MessageableChannel = None,
    ) -> tuple[
        list[equilizer.Equalizer],
        list[equilizer.Equalizer],
        list[equilizer.Equalizer],
        list[equilizer.Equalizer],
        list[equilizer.Equalizer],
    ]:
        """
        Gets all equalizers a user has access to in a given context.

        Globals, User specific, Guild specific, Channel specific, VC specific.

        """
        global_equalizers = [p async for p in self.get_all_equalizers_by_scope(scope=self._client.bot.user.id)]
        user_equalizers = [p async for p in self.get_all_equalizers_by_scope(scope=requester)]
        vc_equalizers = []
        guild_equalizers = []
        channel_equalizers = []
        if voice_channel is not None:
            vc_equalizers = [p async for p in self.get_all_equalizers_by_scope(scope=voice_channel.id)]
        if guild is not None:
            guild_equalizers = [p async for p in self.get_all_equalizers_by_scope(scope=guild.id)]
        if channel is not None:
            channel_equalizers = [p async for p in self.get_all_equalizers_by_scope(scope=channel.id)]
        return global_equalizers, user_equalizers, guild_equalizers, channel_equalizers, vc_equalizers

    async def get_manageable_equalizers(
        self, requester: discord.abc.User, bot: DISCORD_BOT_TYPE, *, name_or_id: str | None = None
    ) -> list[equilizer.Equalizer]:
        if name_or_id:
            try:
                equalizers = await self.get_equalizer_by_name_or_id(name_or_id)
            except EntryNotFoundException:
                equalizers = []
        else:
            try:
                equalizers = [p async for p in self.get_all_equalizers()]
            except EntryNotFoundException:
                equalizers = []
        returning_list = []
        if equalizers:
            async for equalizer in AsyncIter(equalizers):
                if await equalizer.can_manage(requester=requester, bot=bot):
                    returning_list.append(equalizer)
        return returning_list

    @staticmethod
    async def count() -> int:
        """Returns the number of equalizers in the database."""
        return await EqualizerRow.count()



from __future__ import annotations

from math import sqrt

from pylav.constants.misc import ASCII_COLOURS
from pylav.type_hints.generics import SupportsStr


class _BackgroundColourCodes:
    dark_blue = "40"
    orange = "41"
    blue = "42"
    turquoise = "43"
    gray = "44"
    indigo = "45"
    light_gray = "46"
    white = "47"


class EightBitANSI:
    escape = "\u001b["

    black = "30"
    red = "31"
    green = "32"
    yellow = "33"
    blue = "34"
    magenta = "35"
    cyan = "36"
    white = "37"

    normal = reset = "0"
    bold = "1"
    italic = "3"
    underline = "4"
    default = "39"
    background = _BackgroundColourCodes

    @classmethod
    def colorize(
        cls,
        text: SupportsStr,
        color: str,
        bold: bool = False,
        underline: bool = False,
        background: str = None,
        italic: bool = False,
    ) -> str:
        colour = [getattr(cls, color, "39")]
        if background and (background := getattr(cls.background, background, None)):
            colour.append(background)
        if bold:
            colour.append(cls.bold)
        if italic:
            colour.append(cls.italic)
        if underline:
            colour.append(cls.underline)

        colour_code = f"{cls.escape}{';'.join(colour)}m"
        colour_reset = f"{cls.escape}{cls.reset}m"
        text = f"{text}".replace("\n", f"{colour_reset}\n{colour_code}")

        return f"{colour_code}{text}{colour_reset}"

    @classmethod
    def paint_black(
        cls,
        text: SupportsStr,
        bold: bool = False,
        underline: bool = False,
        background: str = None,
        italic: bool = False,
    ) -> str:
        return cls.colorize(text, "black", bold, underline, background, italic)

    @classmethod
    def paint_red(
        cls,
        text: SupportsStr,
        bold: bool = False,
        underline: bool = False,
        background: str = None,
        italic: bool = False,
    ) -> str:
        return cls.colorize(text, "red", bold, underline, background, italic)

    @classmethod
    def paint_green(
        cls,
        text: SupportsStr,
        bold: bool = False,
        underline: bool = False,
        background: str = None,
        italic: bool = False,
    ) -> str:
        return cls.colorize(text, "green", bold, underline, background, italic)

    @classmethod
    def paint_yellow(
        cls,
        text: SupportsStr,
        bold: bool = False,
        underline: bool = False,
        background: str = None,
        italic: bool = False,
    ) -> str:
        return cls.colorize(text, "yellow", bold, underline, background, italic)

    @classmethod
    def paint_blue(
        cls,
        text: SupportsStr,
        bold: bool = False,
        underline: bool = False,
        background: str = None,
        italic: bool = False,
    ) -> str:
        return cls.colorize(text, "blue", bold, underline, background, italic)

    @classmethod
    def paint_magenta(
        cls,
        text: SupportsStr,
        bold: bool = False,
        underline: bool = False,
        background: str = None,
        italic: bool = False,
    ) -> str:
        return cls.colorize(text, "magenta", bold, underline, background, italic)

    @classmethod
    def paint_cyan(
        cls,
        text: SupportsStr,
        bold: bool = False,
        underline: bool = False,
        background: str = None,
        italic: bool = False,
    ) -> str:
        return cls.colorize(text, "cyan", bold, underline, background, italic)

    @classmethod
    def paint_white(
        cls,
        text: SupportsStr,
        bold: bool = False,
        underline: bool = False,
        background: str = None,
        italic: bool = False,
    ) -> str:
        return cls.colorize(text, "white", bold, underline, background, italic)

    @classmethod
    def closest_from_rgb(cls, r: int, g: int, b: int) -> str:
        """Get the closest 4-bit ANSI colour from a given RGB value."""
        return cls.closest_color(r, g, b)

    @classmethod
    def closest_from_hex(cls, value: str) -> str:
        """Get the closest 4-bit ANSI colour from a given hex value."""
        value = value.lstrip("#")
        lv = len(value)
        return cls.closest_color(*tuple(int(value[i : i + lv // 3], 16) for i in range(0, lv, lv // 3)))

    @classmethod
    def closest_color(cls, red: int, green: int, blue: int) -> str:
        color_diffs = []
        for name, color in ASCII_COLOURS.items():
            cr, cg, cb = color
            color_diff = sqrt((red - cr) ** 2 + (green - cg) ** 2 + (blue - cb) ** 2)
            color_diffs.append((color_diff, name))
        return min(color_diffs)[1]



from __future__ import annotations

from collections.abc import Callable, Iterable
from typing import Any

from pylav.storage.database.cache.cache import CACHE


def key_builder(method: Callable, *args: Any, **kwargs: Any) -> str:  # noqa
    return f"{method.__module__}:{args[0].__class__.__name__}:{method.__name__}:{args[0].get_cache_key()}"  # noqa


async def invalidate_cache(method: Callable, instance: object) -> None:
    await CACHE.delete(key=key_builder(method, instance))


async def update_cache(method: Callable, instance: object, value: Any) -> None:
    await CACHE.set(key=key_builder(method, instance), value=value)


async def update_cache_multi(pairs: Iterable[tuple[Callable, Any]], instance: object) -> None:
    await CACHE.set_many(pairs={key_builder(method, instance): value for method, value in pairs})


async def invalidate_cache_multi(methods: Iterable[Callable], instance: object) -> None:
    await CACHE.delete_many(*[key_builder(method, instance) for method in methods])



from __future__ import annotations

from typing import TYPE_CHECKING

from pylav.nodes.api.responses.playlists import Info
from pylav.nodes.api.responses.rest_api import NoMatches
from pylav.nodes.api.responses.websocket import Stats as StatsMessage

if TYPE_CHECKING:
    from pylav.nodes.node import Node


async def sort_key_nodes(node: Node, region: str = None) -> float:
    return await node.penalty_with_region(region)


class Penalty:
    """Represents the penalty of the stats of a Node"""

    __slots__ = ("_stats",)

    def __init__(self, stats: Stats) -> None:
        self._stats = stats

    @property
    def player_penalty(self) -> int:
        """The penalty of the players playing on the node.

        This is the number of players playing in the node.
        """
        return self._stats.playing_players

    @property
    def cpu_penalty(self) -> float:
        """The penalty of the cpu load of the node"""
        return 1.05 ** (100 * self._stats.system_load) * 10 - 10

    @property
    def null_frame_penalty(self) -> float | int:
        """The penalty of the nulled frames of the node"""
        null_frame_penalty = 0
        if self._stats.frames_nulled != -1:
            null_frame_penalty = (1.03 ** (500 * (self._stats.frames_nulled / 3000))) * 300 - 300
            null_frame_penalty *= 2
        return null_frame_penalty

    @property
    def deficit_frame_penalty(self) -> float | int:
        """The penalty of the deficit frames of the node"""
        return (
            1.03 ** (500 * (self._stats.frames_deficit / 3000)) * 600 - 600 if self._stats.frames_deficit != -1 else 0
        )

    @property
    def total(self) -> float:
        """The total penalty of the node.

        This is the sum of the penalties of the node.
        """
        # noinspection PyProtectedMember
        return (
            self.player_penalty
            + self.cpu_penalty
            + self.null_frame_penalty
            + self.deficit_frame_penalty
            + self._stats._node.down_votes * 100
        )

    def __repr__(self) -> str:
        # noinspection PyProtectedMember
        return (
            f"<Penalty player={self.player_penalty} "
            f"cpu={self.cpu_penalty} "
            f"null_frame={self.null_frame_penalty} "
            f"deficit_frame={self.deficit_frame_penalty} "
            f"votes={self._stats._node.down_votes * 100} "
            f"total={self.total}>"
        )


class Stats:
    """Represents the stats of Lavalink node"""

    __slots__ = (
        "_node",
        "_data",
        "_penalty",
        "_memory",
        "_cpu",
        "_frame_stats",
    )

    def __init__(self, node: Node, data: StatsMessage) -> None:
        self._node = node
        self._data = data
        self._memory = data.memory
        self._cpu = data.cpu
        self._frame_stats = data.frameStats
        self._penalty = Penalty(self)

    @property
    def uptime(self) -> int:
        """How long the node has been running for in milliseconds"""
        return self._data.uptime

    @property
    def uptime_seconds(self) -> float:
        """How long the node has been running for in seconds"""
        return self.uptime / 1000

    @property
    def players(self) -> int:
        """The amount of players connected to the node"""
        return self._data.players or self._node.connected_count

    @property
    def playing_players(self) -> int:
        """The amount of players that are playing in the node"""
        return self._data.playingPlayers or self._node.playing_count

    @property
    def memory_free(self) -> int:
        """The amount of memory free to the node"""
        return self._memory.free

    @property
    def memory_used(self) -> int:
        """The amount of memory that is used by the node"""
        return self._memory.used

    @property
    def memory_allocated(self) -> int:
        """The amount of memory allocated to the node"""
        return self._memory.allocated

    @property
    def memory_reservable(self) -> int:
        """The amount of memory reservable to the node"""
        return self._memory.reservable

    @property
    def cpu_cores(self) -> int:
        """The amount of cpu cores the system of the node has"""
        return self._cpu.cores

    @property
    def system_load(self) -> float:
        """The overall CPU load of the system"""
        return self._cpu.systemLoad

    @property
    def lavalink_load(self) -> float:
        """The CPU load generated by Lavalink"""
        return self._cpu.lavalinkLoad

    @property
    def frames_sent(self) -> int:
        """The number of frames sent to Discord.
        Warning
        -------
        Given that audio packets are sent via UDP, this number may not be 100% accurate due to dropped packets.
        """
        return self._frame_stats.sent if self._frame_stats else -1

    @property
    def frames_nulled(self) -> int:
        """The number of frames that yielded null, rather than actual data"""
        return self._frame_stats.nulled if self._frame_stats else -1

    @property
    def frames_deficit(self) -> int:
        """The number of missing frames. Lavalink generates this figure by calculating how many packets to expect
        per minute, and deducting ``frames_sent``. Deficit frames could mean the CPU is overloaded, and isn't
        generating frames as quickly as it should be.
        """
        return self._frame_stats.deficit if self._frame_stats else -1

    @property
    def penalty(self) -> Penalty:
        """The penalty for the node"""
        return self._penalty

    def __repr__(self) -> str:
        return (
            f"<Stats node_id={self._node.identifier} "
            f"uptime={self.uptime} "
            f"players={self.players} "
            f"playing_players={self.playing_players} "
            f"memory_free={self.memory_free} "
            f"memory_used={self.memory_used} "
            f"memory_allocated={self.memory_allocated} "
            f"memory_reservable={self.memory_reservable} "
            f"cpu_cores={self.cpu_cores} "
            f"system_load={self.system_load} "
            f"lavalink_load={self.lavalink_load} "
            f"frames_sent={self.frames_sent} "
            f"frames_nulled={self.frames_nulled} "
            f"frames_deficit={self.frames_deficit}> "
            f"penalty={self.penalty}>"
        )


NO_MATCHES = NoMatches(loadType="NO_MATCHES", tracks=[], playlistInfo=Info(name="", selectedTrack=-1))



from __future__ import annotations

from typing import TYPE_CHECKING

import discord

from pylav.events.base import PyLavEvent

if TYPE_CHECKING:
    from pylav.nodes.api.responses.websocket import Closed
    from pylav.nodes.node import Node
    from pylav.players.player import Player


class NodeDisconnectedEvent(PyLavEvent):
    """This event is dispatched when a node disconnects and becomes unavailable."""

    __slots__ = ()

    def __init__(self, node: Node, code: int, reason: str) -> None:
        self.node = node
        self.code = code
        self.reason = reason


class NodeConnectedEvent(PyLavEvent):
    """This event is dispatched when PyLav successfully connects to a node."""

    __slots__ = ()

    def __init__(self, node: Node) -> None:
        self.node = node


class NodeChangedEvent(PyLavEvent):
    """This event is dispatched when a player changes to another node.
    Keep in mind this event can be dispatched multiple times if a node
    disconnects and the load balancer moves players to a new node.
    """

    __slots__ = ()

    def __init__(self, player: Player, old_node: Node, new_node: Node) -> None:
        self.player = player
        self.old_node = old_node
        self.new_node = new_node


class WebSocketClosedEvent(PyLavEvent):
    """This event is dispatched when an audio websocket to Discord
    is closed. This can happen for various reasons like an
    expired voice server update.
    """

    __slots__ = ()

    def __init__(
        self,
        player: Player,
        node: Node,
        channel: discord.channel.VocalGuildChannel,
        event_object: Closed,
    ) -> None:
        self.player = player
        self.code = event_object.code
        self.reason = event_object.reason
        self.by_remote = event_object.byRemote
        self.node = node
        self.channel = channel
        self.event = event_object



from __future__ import annotations

import dataclasses
from typing import Literal


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class IPBlock:
    type: Literal["Inet4Address", "Inet6Address"]
    size: str


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class FailingAddress:
    address: str
    failingTimestamp: int
    failingTimes: str


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Details:
    ipBlock: IPBlock | dict
    failingAddresses: list[FailingAddress]
    rotateIndex: str
    ipIndex: str
    currentAddress: str
    currentAddressIndex: str
    blockIndex: str

    def __post_init__(self):
        temp = []
        for s in self.failingAddresses:
            if isinstance(s, FailingAddress) or (isinstance(s, dict) and (s := FailingAddress(**s))):
                temp.append(s)
        object.__setattr__(self, "failingAddresses", temp)
        if isinstance(self.ipBlock, dict):
            object.__setattr__(self, "ipBlock", IPBlock(**self.ipBlock))


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Status:
    type: Literal["RotatingIpRoutePlanner", "NanoIpRoutePlanner", "RotatingNanoIpRoutePlanner"] | None = None
    details: Details | None | dict = None

    def __post_init__(self):
        if isinstance(self.details, dict):
            object.__setattr__(self, "details", Details(**self.details))



from __future__ import annotations

from typing import TYPE_CHECKING

from packaging.version import Version

from pylav.constants.versions import VERSION_0_0_0, VERSION_0_9_2_0
from pylav.storage.migrations.logging import LOGGER

if TYPE_CHECKING:
    from pylav.core.client import Client


async def migration_v_0_9_2_0(client: Client, current_version: Version) -> None:
    if current_version >= VERSION_0_9_2_0 or current_version == VERSION_0_0_0:
        return

    LOGGER.info("Running %s migration", VERSION_0_9_2_0)
    await client.player_state_db_manager.delete_all_players()
    await client.lib_db_manager.update_bot_dv_version(VERSION_0_9_2_0)



from __future__ import annotations

from piccolo.columns import BigInt, Float, Text
from piccolo.columns.indexes import IndexMethod
from piccolo.table import Table

from pylav.storage.database.tables.misc import DATABASE_ENGINE


class EqualizerRow(Table, db=DATABASE_ENGINE, tablename="equalizer"):
    id = BigInt(primary_key=True, index=True)
    scope = BigInt(null=True, default=None, index=True)
    name = Text(null=True, default=None, index_method=IndexMethod.gin)
    description = Text(null=True, default=None)
    # FIXME: Drop existing table and recreate it with this column defaults
    # TODO: Handle new defaults
    author = BigInt(null=True, default=None, index=True)
    band_25 = Float(null=True)
    band_40 = Float(null=True)
    band_63 = Float(null=True)
    band_100 = Float(null=True)
    band_160 = Float(null=True)
    band_250 = Float(null=True)
    band_400 = Float(null=True)
    band_630 = Float(null=True)
    band_1000 = Float(null=True)
    band_1600 = Float(null=True)
    band_2500 = Float(null=True)
    band_4000 = Float(null=True)
    band_6300 = Float(null=True)
    band_10000 = Float(null=True)
    band_16000 = Float(null=True)



from __future__ import annotations

import dataclasses
from typing import NotRequired  # noqa

from pylav.constants.node import TRACK_VERSION
from pylav.nodes.api.responses.shared import TrackPluginInfo
from pylav.type_hints.dict_typing import JSON_DICT_TYPE


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Info:
    identifier: str
    isSeekable: bool
    author: str
    length: int = 0
    isStream: bool = False
    position: int | None = 0
    title: str = ""
    uri: str | None = None
    sourceName: str | None = None
    artworkUrl: str | None = None
    isrc: str | None = None
    version: NotRequired[int] = TRACK_VERSION

    def to_dict(self) -> JSON_DICT_TYPE:
        return dataclasses.asdict(self)

    def to_database(self) -> JSON_DICT_TYPE:
        # noinspection SpellCheckingInspection
        return {
            "identifier": self.identifier,
            "title": self.title,
            "uri": self.uri,
            "sourceName": self.sourceName,
            "isrc": self.isrc,
        }


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Track:
    info: Info | dict
    encoded: str | None = None
    pluginInfo: None | dict | TrackPluginInfo = None

    def __post_init__(self) -> None:
        if isinstance(self.pluginInfo, dict):
            object.__setattr__(self, "pluginInfo", TrackPluginInfo(kwargs=self.pluginInfo))
        elif self.pluginInfo is None:
            object.__setattr__(self, "pluginInfo", TrackPluginInfo(kwargs=None))
        if isinstance(self.info, dict):
            object.__setattr__(self, "info", Info(**self.info))

    def set_version(self, version: int) -> None:
        object.__setattr__(self, "version", version)

    def to_dict(self) -> JSON_DICT_TYPE:
        return {
            "info": dataclasses.asdict(self.info),
            "encoded": self.encoded,
            "pluginInfo": dataclasses.asdict(self.pluginInfo) if self.pluginInfo else None,
        }



from __future__ import annotations

from piccolo.columns import Bytea, Text
from piccolo.table import Table

from pylav.storage.database.tables.misc import DATABASE_ENGINE


class AioHttpCacheRow(Table, db=DATABASE_ENGINE, tablename="aiohttp_client_cache"):
    key = Text(primary_key=True, index=True)
    value = Bytea()



from __future__ import annotations

from cashews import Cache  # type: ignore

from pylav.constants.config import READ_CACHING_ENABLED
from pylav.storage.database.cache.logging import LOGGER

if READ_CACHING_ENABLED:
    LOGGER.warning(
        "Caching is enabled, "
        "this will make it so live edits to the database will not be reflected "
        "in the bot until the cache is invalidated or bot is restarted."
    )
else:
    LOGGER.info(
        "Caching is disabled, "
        "this will make it so live edits to the database will be reflected in the bot immediately."
    )


CACHE = Cache("ReadCache")
CACHE.setup("mem://?check_interval=10&size=10000", disable=not READ_CACHING_ENABLED)
# TODO: Allow for redis caching



from __future__ import annotations

import asyncio
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar

import asyncstdlib
from asyncstdlib import heapq
from discord.app_commands import Choice, Transformer
from discord.ext import commands
from rapidfuzz import fuzz

from pylav.exceptions.database import EntryNotFoundException
from pylav.helpers.format.strings import shorten_string
from pylav.storage.models.playlist import Playlist as PlaylistModel
from pylav.type_hints.bot import DISCORD_CONTEXT_TYPE, DISCORD_INTERACTION_TYPE

try:
    from redbot.core.i18n import Translator  # type: ignore

    _ = Translator("PyLav", Path(__file__))
except ImportError:
    Translator = None

    def _(string: str) -> str:
        return string


if TYPE_CHECKING:

    PlaylistConverter = TypeVar("PlaylistConverter", bound=list[PlaylistModel])
else:

    class PlaylistConverter(Transformer):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> list[PlaylistModel]:
            """Converts a playlist name or ID to a list of matching objects"""

            try:
                playlists = await ctx.pylav.playlist_db_manager.get_playlist_by_name_or_id(arg)
            except EntryNotFoundException as e:
                raise commands.BadArgument(_("Playlist with name or id `{arg}` not found").format(arg=arg)) from e
            return playlists

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> list[PlaylistModel]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:

            if not current:
                playlists = await interaction.client.pylav.playlist_db_manager.get_bundled_playlists()
                return [
                    Choice(name=shorten_string(await e.fetch_name(), max_length=100), value=f"{e.id}")
                    for e in playlists
                ][:25]

            try:
                playlists = await interaction.client.pylav.playlist_db_manager.get_playlist_by_name(current, limit=50)
            except EntryNotFoundException:
                return []

            async def _filter(c: PlaylistModel):
                name = await c.fetch_name()
                author = await c.fetch_author()
                return (
                    await asyncio.to_thread(fuzz.partial_ratio, name, current, score_cutoff=75),
                    1 if author == interaction.user.id else 0,
                    [-ord(i) for i in name],
                )

            extracted = await heapq.nlargest(asyncstdlib.iter(playlists), n=25, key=_filter)
            return [
                Choice(name=shorten_string(await e.fetch_name(), max_length=100), value=f"{e.id}") for e in extracted
            ]



from __future__ import annotations

from pylav.players.filters.misc import FilterMixin


class ChannelMix(FilterMixin):
    __slots__ = ("_left_to_left", "_left_to_right", "_right_to_left", "_right_to_right", "_default")

    def __init__(
        self,
        left_to_left: float | None = None,
        left_to_right: float | None = None,
        right_to_left: float | None = None,
        right_to_right: float | None = None,
    ) -> None:
        super().__init__()
        self.left_to_left = left_to_left
        self.left_to_right = left_to_right
        self.right_to_left = right_to_left
        self.right_to_right = right_to_right

    def to_dict(self) -> dict[str, float | None | bool]:
        return {
            "leftToLeft": self.left_to_left,
            "leftToRight": self.left_to_right,
            "rightToLeft": self.right_to_left,
            "rightToRight": self.right_to_right,
        }

    @classmethod
    def from_dict(cls, data: dict[str, float | bool | None]) -> ChannelMix:
        return cls(
            left_to_left=data["leftToLeft"],
            left_to_right=data["leftToRight"],
            right_to_left=data["rightToLeft"],
            right_to_right=data["rightToRight"],
        )

    def __repr__(self) -> str:
        return (
            f"<ChannelMix: left_to_left={self.left_to_left}, "
            f"left_to_right={self.left_to_right}, "
            f"right_to_left={self.right_to_left}, "
            f"right_to_right={self.right_to_right}>"
        )

    @property
    def left_to_left(self) -> float | None:
        return self._left_to_left

    @left_to_left.setter
    def left_to_left(self, v: float | None) -> None:
        self._left_to_left = v

    @property
    def left_to_right(self) -> float | None:
        return self._left_to_right

    @left_to_right.setter
    def left_to_right(self, v: float | None) -> None:
        self._left_to_right = v

    @property
    def right_to_left(self) -> float | None:
        return self._right_to_left

    @right_to_left.setter
    def right_to_left(self, v: float | None) -> None:
        self._right_to_left = v

    @property
    def right_to_right(self) -> float | None:
        return self._right_to_right

    @right_to_right.setter
    def right_to_right(self, v: float | None) -> None:
        self._right_to_right = v

    @classmethod
    def default(cls) -> ChannelMix:
        return cls()

    def get(self) -> dict[str, float]:
        if self.off:
            return {}
        response = {}
        if self.left_to_left is not None:
            response["leftToLeft"] = self.left_to_left
        if self.left_to_right is not None:
            response["leftToRight"] = self.left_to_right
        if self.right_to_left is not None:
            response["rightToLeft"] = self.right_to_left
        if self.right_to_right is not None:
            response["rightToRight"] = self.right_to_right
        return response

    def reset(self) -> None:
        self.right_to_right = self.right_to_left = self.left_to_right = self.left_to_left = None



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar, Union

import discord

if TYPE_CHECKING:
    from discord.ext.commands import AutoShardedBot, Cog, CommandError, Context

    from pylav.core.client import Client
    from pylav.core.context import PyLavContext

    try:
        from redbot.core.bot import Red
        from redbot.core.bot import Red as BotClient
        from redbot.core.commands import Cog as RedCog
        from redbot.core.commands import Context as ClientContext
    except ImportError:
        BotClient = Red = AutoShardedBot
        RedCog = Cog
        ClientContext = Context


else:
    from discord.ext.commands import Cog, CommandError

    try:
        from redbot.core.bot import Red as BotClient
        from redbot.core.commands import Cog as RedCog
    except ImportError:
        from discord.ext.commands import AutoShardedBot as BotClient

        RedCog = Cog

from discord import app_commands


class BotClientWithLavalinkType(BotClient):
    _pylav_client: Client
    lavalink: Client
    pylav: Client

    async def get_context(
        self, message: discord.abc.Message | DISCORD_INTERACTION_TYPE | PyLavContext, *, cls: type[PyLavContext] = None
    ) -> PyLavContext[Any]:
        ...


class DISCORD_INTERACTION_TYPE_BASE(discord.Interaction):
    client: BotClientWithLavalinkType
    response: discord.InteractionResponse
    followup: discord.Webhook
    command: app_commands.Command[Any, ..., Any] | app_commands.ContextMenu | None
    channel: discord.interactions.InteractionChannel | None


class DISCORD_COG_TYPE_MIXIN(RedCog):
    __version__: str
    bot: DISCORD_BOT_TYPE
    lavalink: Client
    pylav: Client


DISCORD_BOT_TYPE = TypeVar("DISCORD_BOT_TYPE", bound=BotClientWithLavalinkType, covariant=True)
DISCORD_CONTEXT_TYPE = TypeVar("DISCORD_CONTEXT_TYPE", bound="PyLavContext", covariant=True)
DISCORD_INTERACTION_TYPE = TypeVar(
    "DISCORD_INTERACTION_TYPE", bound=DISCORD_INTERACTION_TYPE_BASE | discord.Interaction, covariant=True
)
DISCORD_COG_TYPE = TypeVar("DISCORD_COG_TYPE", bound=DISCORD_COG_TYPE_MIXIN, covariant=True)
DISCORD_COMMAND_ERROR_TYPE = TypeVar(
    "DISCORD_COMMAND_ERROR_TYPE", bound=Union[CommandError, app_commands.errors.AppCommandError], covariant=True
)



from __future__ import annotations

import asyncio
import contextlib
import datetime
from typing import TYPE_CHECKING

import asyncpg

from pylav.helpers.time import get_now_utc
from pylav.logging import getLogger
from pylav.players.query.obj import Query as QueryObj
from pylav.storage.database.tables.queries import QueryRow
from pylav.storage.database.tables.tracks import TrackRow
from pylav.storage.models.query import Query
from pylav.type_hints.dict_typing import JSON_DICT_TYPE
from pylav.utils.vendor.redbot import AsyncIter

if TYPE_CHECKING:
    from pylav.core.client import Client
LOGGER = getLogger("PyLav.Database.Controller.Query")


class QueryController:
    __slots__ = ("_client",)

    def __init__(self, client: Client) -> None:
        self._client = client

    @property
    def client(self) -> Client:
        return self._client

    @staticmethod
    async def exists(query: QueryObj) -> bool:
        return await QueryRow.exists().where(
            (QueryRow.identifier == query.query_identifier)
            & (QueryRow.last_updated > get_now_utc() - datetime.timedelta(days=30))
        )

    @staticmethod
    def get(identifier: str) -> Query:
        """Get a query object"""
        return Query(id=identifier)

    async def fetch_query(self, query: QueryObj) -> Query | None:
        if query.is_local or query.is_custom_playlist or query.is_http:
            # Do not cache local queries and single track urls or http source entries
            return None

        if await self.exists(query):
            cached = self.get(query.query_identifier)
            if await cached.size() > 0:
                return cached

    @staticmethod
    async def add_query(query: QueryObj, result: JSON_DICT_TYPE) -> bool:
        if query.is_local or query.is_custom_playlist or query.is_http:
            # Do not cache local queries and single track urls or http source entries
            return False
        if result.get("loadType") in ["NO_MATCHES", "LOAD_FAILED", None]:
            return False
        tracks = result.get("tracks", [])
        if not tracks:
            return False
        playlist_info = result.get("playlistInfo", {})
        name = playlist_info.get("name", None) if playlist_info else None
        defaults = {QueryRow.name: name}
        query_row = await QueryRow.objects().get_or_create(QueryRow.identifier == query.query_identifier, defaults)

        # noinspection PyProtectedMember
        if not query_row._was_created:
            await QueryRow.update(defaults).where(QueryRow.identifier == query.query_identifier)
        new_tracks = []
        # TODO: Optimize this, after https://github.com/piccolo-orm/piccolo/discussions/683 is answered or fixed
        async for track in AsyncIter(tracks):
            with contextlib.suppress(Exception):
                for key in track["info"].keys():
                    # noinspection SpellCheckingInspection
                    if key not in {"identifier", "sourceName", "title", "uri", "isrc"}:
                        track["info"].pop(key, None)
                new_tracks.append(await TrackRow.get_or_create(track["encoded"], track["info"]))
        if new_tracks:
            await query_row.add_m2m(*new_tracks, m2m=QueryRow.tracks)
            return True
        await QueryRow.delete().where(QueryRow.identifier == query.query_identifier)
        return False

    @staticmethod
    async def delete_old() -> None:
        with contextlib.suppress(asyncio.exceptions.CancelledError, asyncpg.exceptions.CannotConnectNowError):
            LOGGER.trace("Deleting old queries")
            await QueryRow.delete().where(QueryRow.last_updated <= (get_now_utc() - datetime.timedelta(days=30)))
            LOGGER.trace("Deleted old queries")

    @staticmethod
    async def wipe() -> None:
        LOGGER.trace("Wiping query cache")
        await QueryRow.raw(
            "TRUNCATE TABLE query",
        )
        LOGGER.trace("Wiped query cache")

    @staticmethod
    async def delete_older_than(days: int) -> None:
        await QueryRow.delete().where(QueryRow.last_updated <= (get_now_utc() - datetime.timedelta(days=days)))

    @staticmethod
    async def delete_query(query: QueryObj) -> None:
        await QueryRow.delete().where(QueryRow.identifier == query.query_identifier)

    @staticmethod
    async def size() -> int:
        return await QueryRow.count()



from __future__ import annotations

import os
import pathlib
import sys

import aiopath  # type: ignore
import platformdirs

# noinspection PyProtectedMember
from pylav._internals.pylav_yaml_builder import build_from_envvars
from pylav.logging import getLogger

LOGGER = getLogger("PyLav.Environment")

INSTANCE_NAME = None

LIB_DIR = platformdirs.PlatformDirs("PyLav")
__CONFIG_DIR = pathlib.Path(LIB_DIR.user_config_path)
if sys.platform == "linux" and 0 < os.getuid() < 1000 and not pathlib.Path.home().exists():
    __CONFIG_DIR = pathlib.Path(LIB_DIR.site_data_path)
__CONFIG_DIR.mkdir(parents=True, exist_ok=True)
CONFIG_DIR = aiopath.AsyncPath(__CONFIG_DIR)


ENV_FILE: pathlib.Path = pathlib.Path(os.getenv("PYLAV__YAML_CONFIG", pathlib.Path.home() / "pylav.yaml"))


if not ENV_FILE.exists():
    LOGGER.warning(
        "%s does not exist - This is not a problem if it does then the environment variables will be read from it",
        ENV_FILE,
    )
    build_from_envvars()
    from pylav.constants.config.env_var import DEFAULT_SEARCH_SOURCE as DEFAULT_SEARCH_SOURCE
    from pylav.constants.config.env_var import EXTERNAL_UNMANAGED_HOST as EXTERNAL_UNMANAGED_HOST
    from pylav.constants.config.env_var import EXTERNAL_UNMANAGED_PASSWORD as EXTERNAL_UNMANAGED_PASSWORD
    from pylav.constants.config.env_var import EXTERNAL_UNMANAGED_PORT as EXTERNAL_UNMANAGED_PORT
    from pylav.constants.config.env_var import EXTERNAL_UNMANAGED_SSL as EXTERNAL_UNMANAGED_SSL
    from pylav.constants.config.env_var import FALLBACK_POSTGREST_HOST as FALLBACK_POSTGREST_HOST
    from pylav.constants.config.env_var import JAVA_EXECUTABLE as JAVA_EXECUTABLE
    from pylav.constants.config.env_var import LINKED_BOT_IDS as LINKED_BOT_IDS
    from pylav.constants.config.env_var import MANAGED_NODE_APPLE_MUSIC_API_KEY as MANAGED_NODE_APPLE_MUSIC_API_KEY
    from pylav.constants.config.env_var import (
        MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE as MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE,
    )
    from pylav.constants.config.env_var import MANAGED_NODE_DEEZER_KEY as MANAGED_NODE_DEEZER_KEY
    from pylav.constants.config.env_var import MANAGED_NODE_SPOTIFY_CLIENT_ID as MANAGED_NODE_SPOTIFY_CLIENT_ID
    from pylav.constants.config.env_var import MANAGED_NODE_SPOTIFY_CLIENT_SECRET as MANAGED_NODE_SPOTIFY_CLIENT_SECRET
    from pylav.constants.config.env_var import MANAGED_NODE_SPOTIFY_COUNTRY_CODE as MANAGED_NODE_SPOTIFY_COUNTRY_CODE
    from pylav.constants.config.env_var import (
        MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN as MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN,
    )
    from pylav.constants.config.env_var import POSTGRES_DATABASE as POSTGRES_DATABASE
    from pylav.constants.config.env_var import POSTGRES_HOST as POSTGRES_HOST
    from pylav.constants.config.env_var import POSTGRES_PASSWORD as POSTGRES_PASSWORD
    from pylav.constants.config.env_var import POSTGRES_PORT as POSTGRES_PORT
    from pylav.constants.config.env_var import POSTGRES_SOCKET as POSTGRES_SOCKET
    from pylav.constants.config.env_var import POSTGRES_USER as POSTGRES_USER
    from pylav.constants.config.env_var import READ_CACHING_ENABLED as READ_CACHING_ENABLED
    from pylav.constants.config.env_var import REDIS_FULL_ADDRESS_RESPONSE_CACHE as REDIS_FULL_ADDRESS_RESPONSE_CACHE
    from pylav.constants.config.env_var import (
        TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS as TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS,
    )
    from pylav.constants.config.env_var import (
        TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS as TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS,
    )
    from pylav.constants.config.env_var import (
        TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS as TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS,
    )
    from pylav.constants.config.env_var import (
        USE_BUNDLED_EXTERNAL_LAVA_LINK_NODE as USE_BUNDLED_EXTERNAL_LAVA_LINK_NODE,
    )
    from pylav.constants.config.env_var import USE_BUNDLED_EXTERNAL_PYLAV_NODE as USE_BUNDLED_EXTERNAL_PYLAV_NODE
else:
    LOGGER.info("%s exist - Environment variables will be read from it", ENV_FILE)
    from pylav.constants.config.file import DEFAULT_SEARCH_SOURCE as DEFAULT_SEARCH_SOURCE
    from pylav.constants.config.file import EXTERNAL_UNMANAGED_HOST as EXTERNAL_UNMANAGED_HOST
    from pylav.constants.config.file import EXTERNAL_UNMANAGED_PASSWORD as EXTERNAL_UNMANAGED_PASSWORD
    from pylav.constants.config.file import EXTERNAL_UNMANAGED_PORT as EXTERNAL_UNMANAGED_PORT
    from pylav.constants.config.file import EXTERNAL_UNMANAGED_SSL as EXTERNAL_UNMANAGED_SSL
    from pylav.constants.config.file import FALLBACK_POSTGREST_HOST as FALLBACK_POSTGREST_HOST
    from pylav.constants.config.file import JAVA_EXECUTABLE as JAVA_EXECUTABLE
    from pylav.constants.config.file import LINKED_BOT_IDS as LINKED_BOT_IDS
    from pylav.constants.config.file import MANAGED_NODE_APPLE_MUSIC_API_KEY as MANAGED_NODE_APPLE_MUSIC_API_KEY
    from pylav.constants.config.file import (
        MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE as MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE,
    )
    from pylav.constants.config.file import MANAGED_NODE_DEEZER_KEY as MANAGED_NODE_DEEZER_KEY
    from pylav.constants.config.file import MANAGED_NODE_SPOTIFY_CLIENT_ID as MANAGED_NODE_SPOTIFY_CLIENT_ID
    from pylav.constants.config.file import MANAGED_NODE_SPOTIFY_CLIENT_SECRET as MANAGED_NODE_SPOTIFY_CLIENT_SECRET
    from pylav.constants.config.file import MANAGED_NODE_SPOTIFY_COUNTRY_CODE as MANAGED_NODE_SPOTIFY_COUNTRY_CODE
    from pylav.constants.config.file import (
        MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN as MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN,
    )
    from pylav.constants.config.file import POSTGRES_DATABASE as POSTGRES_DATABASE
    from pylav.constants.config.file import POSTGRES_HOST as POSTGRES_HOST
    from pylav.constants.config.file import POSTGRES_PASSWORD as POSTGRES_PASSWORD
    from pylav.constants.config.file import POSTGRES_PORT as POSTGRES_PORT
    from pylav.constants.config.file import POSTGRES_SOCKET as POSTGRES_SOCKET
    from pylav.constants.config.file import POSTGRES_USER as POSTGRES_USER
    from pylav.constants.config.file import READ_CACHING_ENABLED as READ_CACHING_ENABLED
    from pylav.constants.config.file import REDIS_FULL_ADDRESS_RESPONSE_CACHE as REDIS_FULL_ADDRESS_RESPONSE_CACHE
    from pylav.constants.config.file import (
        TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS as TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS,
    )
    from pylav.constants.config.file import (
        TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS as TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS,
    )
    from pylav.constants.config.file import (
        TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS as TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS,
    )
    from pylav.constants.config.file import USE_BUNDLED_EXTERNAL_LAVA_LINK_NODE as USE_BUNDLED_EXTERNAL_LAVA_LINK_NODE
    from pylav.constants.config.file import USE_BUNDLED_EXTERNAL_PYLAV_NODE as USE_BUNDLED_EXTERNAL_PYLAV_NODE

BROTLI_ENABLED = False



from __future__ import annotations

from abc import abstractmethod
from typing import Any

from deepdiff import DeepDiff  # type: ignore

from pylav.logging import getLogger

LOGGER = getLogger("PyLav.Filters")


class FilterMixin:
    __slots__ = ("_default", "_default_value")

    def __init__(self) -> None:
        self._default: FilterMixin | None = None
        self._default_value = None

    def __eq__(self, other: Any) -> bool:
        """Overrides the default implementation"""
        if isinstance(other, self.__class__):
            return bool(
                DeepDiff(
                    self.to_dict(),
                    other.to_dict(),
                    ignore_order=True,
                    max_passes=1,
                    cache_size=100,
                    exclude_paths=["root['name']"],
                )
            )
        return NotImplemented

    @abstractmethod
    def to_dict(self) -> dict[str, Any]:
        """Returns a dictionary representation of the Filter"""
        raise NotImplementedError

    def __hash__(self) -> int:
        """Overrides the default implementation"""
        return hash(tuple(sorted(self.to_dict().items())))

    def __bool__(self) -> bool:
        return self.changed

    @property
    def off(self) -> bool:
        return not self.changed

    @property
    def changed(self) -> bool:
        if self._default is None:
            self._default = self.default()

        assert self._default is not None

        changed = DeepDiff(
            self.to_dict(),
            self._default.to_dict(),
            ignore_order=True,
            max_passes=1,
            cache_size=100,
            exclude_paths=["root['name']"],
        )
        return bool(changed)

    @classmethod
    def default(cls) -> FilterMixin:
        return cls()
